#!/usr/bin/env python
"""main"""

import sys

from cutfinder.cutfinder import CutFinder
from cutfinder.utils.utils import parse_arg


def main():
    """entry point for command line execution"""
    CutFinder(*map(parse_arg, sys.argv[1:]))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
a function to provide lumber cutting schemes optimized for maximal size of remaining pieces

"""
__author__ = "ccluff"

from cutfinder.utils.board_sets import StockBoardSet, FinalBoardSet
from cutfinder.utils.boards import StockBoard, FinalBoard
from cutfinder.utils.utils import parse_arg

try:
    from collections.abc import Iterable  # noqa
except ImportError:
    from collections import Iterable

from itertools import combinations
from functools import partial
from typing import Tuple, List


KERF = 0.125


class CutFinder:
    """Find best way to cut a group of final boards from a group of stock boards.
    Notes:
        -Including more than 15 stock boards or final boards could result in
    slow runtime as the number of possible combinations explodes
        -Best is defined as "leaving the largest pieces", i.e. leaving 10 is better than leaving 2x5"""

    def __init__(self, stocks: List, finals: List, kerf: float = KERF):
        self.stocks = StockBoardSet(stocks)
        self.finals = FinalBoardSet(finals)
        self.kerf: float = kerf

        self.remove_oversize_finals()

        while self.finals.unaddressed_boards and self.stocks.unaddressed_boards:
            find_lowest_waste = partial(
                self._find_lowest_waste, combos=list(self.all_combos_of_finals)
            )
            best = min(map(find_lowest_waste, self.stocks.unaddressed_boards))
            self.address(best)
        print(self)

    @property
    def all_combos_of_finals(self) -> Iterable:
        for i in range(1, len(self.finals.unaddressed_boards) + 1):
            yield from combinations(self.finals.unaddressed_boards, i)

    def _find_lowest_waste(self, stock: StockBoard, combos: List):
        best = (stock.length, tuple(), stock)
        for possible_group in combos:
            length_of_group = (
                sum(possible_group) + (len(possible_group) - 1) * self.kerf
            )

            if (
                stock.length >= length_of_group
                and stock.length - length_of_group < best[0]
            ):
                remainder = stock - length_of_group
                best = (remainder, possible_group, stock)
        return best

    def address(self, tup: tuple):
        """address final boards to stock board"""
        remainder: float = tup[0]
        finals: Tuple[FinalBoard] = tup[1]
        stock: StockBoard = tup[2]

        stock.remainder = max([0.0, remainder - self.kerf])
        stock.used = True
        stock.addressed = True
        stock.cut_into.extend([board.length for board in finals])

        for final in finals:
            self.finals.boards[final.id].cut_into = stock.id
            self.finals.boards[final.id].used = True
            self.finals.boards[final.id].addressed = True

    def remove_oversize_finals(self):
        """address any finals larger than the stock, we can't do anything for them"""
        longest_stock = max(self.stocks.boards.values())
        for id_, final in self.finals.boards.items():
            if final > longest_stock:
                self.finals.boards[id_].addressed = True

    def __repr__(self):
        out = ""
        if self.finals.unused_boards:
            out += f"Unable to allocate final cuts: {', '.join(map(str, self.finals.unused_boards))}\n"
        for board in sorted(
            self.stocks.boards.values(), key=lambda y: y.remainder, reverse=True
        ):
            out += f"Stock Board of length {board.length} is "
            if board.cut_into:
                out += f"cut into final boards {', '.join(map(str, board.cut_into))}"
            else:
                out += f"uncut"
            out += f" with remainder {board.remainder}\n"

        return out


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1:
        CutFinder(*map(parse_arg, sys.argv[1:]))



#!/usr/bin/env python
"""
A new .py file

"""

__author__ = 'ccluff'

from cutfinder.utils.boards import StockBoard, FinalBoard


class BoardSet:
    boards = dict()

    def __iter__(self):
        yield from self.boards

    def __getitem__(self, item):
        return self.boards.get(item)

    def __eq__(self, other):
        return [board.length for board in self.boards.values()] == other

    def __len__(self):
        return len(self.boards)

    @property
    def used_boards(self):
        """boards that haven't been allocated"""
        return [board.length for board in self.boards.values() if board.used]

    @property
    def unused_boards(self):
        """boards that haven't been allocated"""
        return [board.length for board in self.boards.values() if not board.used]

    @property
    def unaddressed_boards(self):
        """boards that need to be considered still"""
        return [board for board in self.boards.values() if not board.addressed]


class FinalBoardSet(BoardSet):
    """Final Boards"""

    def __init__(self, dimensions_set):
        if str(dimensions_set).isnumeric():
            dimensions_set = [dimensions_set]
        self.boards = {id_ + 1: FinalBoard(id_ + 1, dim) for id_, dim in enumerate(dimensions_set)}


class StockBoardSet(BoardSet):
    """Stock Boards"""

    def __init__(self, dimensions_set):
        if str(dimensions_set).isnumeric():
            dimensions_set = [dimensions_set]
        self.boards = {id_ + 1: StockBoard(id_ + 1, dim) for id_, dim in enumerate(dimensions_set)}



#!/usr/bin/env python
"""
A new .py file

"""

__author__ = 'ccluff'

from typing import List


def parse_arg(arg: str) -> List[float]:
    """for parsing cli args from str to python objects"""
    arg = arg.replace('[', '').replace(']', '').replace(' ', '').split(',')
    return list(map(float, arg))



import unittest

from cutfinder.cutfinder import CutFinder


class MainTests(unittest.TestCase):
    def test_basic(self):
        cf = CutFinder(finals=[50, 49], stocks=[100])
        self.assertEqual(cf.finals.used_boards, [50, 49])
        self.assertEqual(cf.finals.unused_boards, list())
        self.assertEqual(cf.stocks.unused_boards, list())

    def test_largest_remainder(self):
        cf = CutFinder(finals=[50], stocks=[100, 200])
        self.assertEqual(cf.finals.used_boards, [50])
        self.assertEqual(cf.finals.unused_boards, list())
        self.assertEqual(cf.stocks.used_boards, [100])
        self.assertEqual(cf.stocks.unused_boards, [200])
        # using the 100 stock leaves the 200, i.e. a larger remainder

    def test_prioritizes_least_waste(self):
        cf = CutFinder(finals=[50, 49, 23], stocks=[100])
        self.assertEqual(cf.finals.unused_boards, [23])
        self.assertEqual(cf.finals.used_boards, [50, 49])
        self.assertEqual(cf.stocks.unused_boards, list())
        # [50, 49] leaves less waste than [50, 23] or [49, 23]

    def test_final_longer_than_stock(self):
        cf = CutFinder(finals=[100], stocks=[99])
        self.assertEqual(cf.finals.unused_boards, [100])
        self.assertEqual(cf.finals.used_boards, list())
        self.assertEqual(cf.stocks.unused_boards, [99])
        self.assertEqual(cf.stocks.used_boards, list())
        # nothing is possible here

    def test_single_kerf(self):
        cf = CutFinder(finals=[99.5], stocks=[100], kerf=1)
        self.assertEqual(cf.finals.unused_boards, list())
        self.assertEqual(cf.finals.used_boards, [99.5])
        self.assertEqual(cf.stocks.unused_boards, list())
        self.assertEqual(cf.stocks.used_boards, [100])
        # you can cut less than a single kerf off 1

    def test_kerf(self):
        cf = CutFinder(finals=[49.5, 49.5], stocks=[100], kerf=1)
        self.assertEqual(cf.finals.unused_boards, list())
        self.assertEqual(cf.finals.used_boards, [49.5, 49.5])
        self.assertEqual(cf.stocks.unused_boards, list())
        self.assertEqual(cf.stocks.used_boards, [100])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
A new .py file

"""

class Board:
    def __init__(self, final_id: int, length: float):
        self.length = length
        self.id = final_id
        self.used = False
        self.addressed = False

    def __add__(self, other):
        if isinstance(other, Board):
            return self.length + other.length
        return self.length + other

    def __sub__(self, other):
        return self.length - other

    def __radd__(self, other):
        return self.length if other == 0 else self.__add__(other)

    def __gt__(self, other):
        if isinstance(other, Board):
            return self.length > other.length
        else:
            return self.length > other

    def __ge__(self, other):
        if isinstance(other, Board):
            return self.length >= other.length
        else:
            return self.length >= other

    def __le__(self, other):
        if isinstance(other, Board):
            return self.length <= other.length
        else:
            return self.length <= other

    def __lt__(self, other):
        if isinstance(other, Board):
            return self.length < other.length
        else:
            return self.length < other

    def __mul__(self, other):
        if isinstance(other, Board):
            return self.length * other.length
        else:
            return self.length * other


class FinalBoard(Board):
    def __init__(self, final_id: int, length: float):
        super().__init__(final_id, length)
        self.source = None

    def __repr__(self):
        if self.source:
            return f"{self.length} from board {self.source}"
        else:
            return str(self.length)


class StockBoard(Board):
    def __init__(self, stock_id: int, length: float):
        super().__init__(stock_id, length)
        self.cut_into = list()
        self.remainder = length

    def __repr__(self):
        return f"{self.length} remaining"


